@@ -3,31 +3,33 @@
 """Module with general utility methods used in tag maps package.
 """
 
-import os
-import sys
-import emoji
-import unicodedata
-import math
+# delay evaluation of annotations at runtime (PEP 563)
+from __future__ import absolute_import, annotations
+
 import argparse
-import numpy as np
-import re
 import hashlib
 import io
 import logging
+import math
+import os
+import platform
+import re
+import sys
+import unicodedata
+from collections import namedtuple
+from datetime import timedelta
+from importlib import reload
+from math import asin, cos, radians, sin, sqrt
+from pathlib import Path
+from typing import Dict, Iterable, List, Optional, Set, Tuple, Union
+
+import emoji
+import numpy as np
 import pyproj
-import fiona
 import regex
 import shapely.geometry as geometry
-import matplotlib.pyplot as plt
-from fiona.crs import from_epsg
-from pathlib import Path
-from shapely.ops import transform, cascaded_union, polygonize
-from datetime import timedelta
-from typing import List, Set, Dict, Tuple, Optional, TextIO, Iterable
-from math import radians, cos, sin, asin, sqrt
-from descartes import PolygonPatch
-from tagmaps.config.config import BaseConfig
-from tagmaps.classes.shared_structure import CleanedPost, AnalysisBounds
+
+from ..classes.shared_structure import AnalysisBounds, ConfigMap, ItemCounter
 
 
 class Utils():
@@ -37,7 +39,92 @@ class Utils():
     """
 
     @staticmethod
-    def _get_shapely_bounds(
+    def check_folder_file(
+            folder_file: Path, create_folder: bool = None) -> Path:
+        """Check if folder exists, optionally create it"""
+        if not folder_file.exists():
+            pname = "File"
+            if folder_file.is_dir():
+                return folder_file
+            pname = "Folder"
+            if create_folder:
+                Utils.init_dir(folder_file)
+                return folder_file
+            raise ValueError(f"{pname} {folder_file} not found.")
+        return folder_file
+
+    @staticmethod
+    def check_fileheader(
+            fieldnames: Iterable[str],
+            source_map: ConfigMap, filename: Optional[str]):
+        """Checks against existing file columns
+        warns if required keys are missing
+
+        Args:
+            source_map (ConfigMap): Custom headers definition
+            Note: the type hint above makes use of forward-references
+            defined in PEP 484, as a means to avoid cyclic imports
+        """
+        header_req = [
+            source_map.post_guid_col,
+            source_map.user_guid_col,
+            source_map.latitude_col,
+            source_map.longitude_col,
+            source_map.tags_col]
+        for header in header_req:
+            if header not in fieldnames:
+                raise Warning(
+                    f'File header is missing "{header}"-column, '
+                    f'file: {filename}')
+
+    @staticmethod
+    def _count_none(str_list: Union[Dict[str, str], List[str]]) -> int:
+        """Return count of list, returns zero for None"""
+        if str_list is None:
+            return 0
+        return len(str_list)
+
+    @staticmethod
+    def report_listload(list_ref, str_text):
+        """Report only if list_ref contains at least 1 entry."""
+        entry_count = Utils._count_none(list_ref)
+        if entry_count > 0:
+            logging.getLogger("tagmaps").info(
+                f"Loaded {entry_count} {str_text}.")
+
+    @staticmethod
+    def concat_topic(term_list):
+        """Concatenate list of terms (e.g. TOPIC) to string"""
+        if any('-' in s for s in term_list):
+            raise ValueError(
+                "No '-' characters supported in topic list terms")
+        topic_name = '-'.join(term_list)
+        return topic_name
+
+    @staticmethod
+    def split_topic(term_concat):
+        """Split concat topic"""
+        topic_terms = term_concat.split('-')
+        return topic_terms
+
+    @staticmethod
+    def set_proj_dir():
+        """Update PROJ_LIB location if not found
+
+        Leftover from pyproj < 2.0.0 compatibility,
+        PROJ_LIB not needed anymore
+        """
+        if not os.environ.get('PROJ_LIB'):
+            local_proj_path = Path.cwd() / "proj"
+            if not local_proj_path.exists():
+                raise ValueError("Pyproj 'proj' datadir not found. Either specify "
+                                 "PROJ_LIB environmental variable or copy 'proj' "
+                                 "folder to local path of executable")
+            os.environ['PROJ_LIB'] = str(local_proj_path)
+            pyproj.datadir.set_data_dir(str(local_proj_path))
+
+    @staticmethod
+    def get_shapely_bounds(
             bounds: AnalysisBounds) -> geometry.MultiPoint:
         """Returns boundary shape from 4 coordinates"""
         bound_points_shapely = geometry.MultiPoint([
@@ -47,7 +134,7 @@ class Utils():
         return bound_points_shapely
 
     @staticmethod
-    def _get_best_utmzone(
+    def get_best_utmzone(
             bound_points_shapely: geometry.MultiPoint):
         """Calculate best UTM Zone SRID/EPSG Code
         Args:
@@ -56,7 +143,7 @@ class Utils():
         input_lat_center = bound_points_shapely.centroid.coords[0][1]
         epsg_code = Utils._convert_wgs_to_utm(
             input_lon_center, input_lat_center)
-        crs_proj = pyproj.Proj(init=f'epsg:{epsg_code}')
+        crs_proj = f'epsg:{epsg_code}'
         return crs_proj, epsg_code
 
     @staticmethod
@@ -87,47 +174,120 @@ class Utils():
         return epsg_code
 
     @staticmethod
-    def default_empty_cstructure():
-        """Generates a tuple of parametric length with
-        empty strings:
-        (" "," "," "," "," "," "," "," "," "," "," "," ")
-        """
-        empty_string_list = list()
-        for _ in range(len(CleanedPost._fields)):
-            empty_string_list.append(" ")
-        empty_string_tuple = tuple(empty_string_list)
-        return empty_string_tuple
-
-    @staticmethod
-    def encode_string(s):
+    def encode_string(text_s):
         """Encode string in Sha256,
         produce hex
 
         - returns a string of double length,
         containing only hexadecimal digits"""
         encoded_string = hashlib.sha3_256(
-            s.encode("utf8")).hexdigest()
+            text_s.encode("utf8")).hexdigest()
         return encoded_string
 
     @staticmethod
-    def _remove_special_chars(s):
+    def remove_special_chars(text_s):
         """Removes a list of special chars from string"""
-        SPECIAL_CHARS = "?.!/;:,[]()'-&#"
-        s_cleaned = s.translate(
-            {ord(c): " " for c in SPECIAL_CHARS})
+        special_chars = "?.!/;:,[]()'-&#|<>=\""
+        s_cleaned = text_s.translate(
+            {ord(c): " " for c in special_chars})
         return s_cleaned
 
     @staticmethod
-    def _is_number(s):
+    def select_words(text_s, selection_list: List[str]) -> str:
+        """Filters a string based on a provided
+        positive filter list of terms
+
+        - removes duplicate terms
+        """
+        # first remove hyperlinks
+        text_s = Utils.remove_hyperlinks(text_s)
+        # split string by space character into list
+        querywords = text_s.split()
+        resultwords = {word for word in querywords if word.lower()
+                       in selection_list}
+        s_cleaned = ' '.join(resultwords)
+        return s_cleaned
+
+    @staticmethod
+    def select_emoji(
+            input_emoji_set: Set[str],
+            selection_emoji_set: Set[str] = None) -> Set[str]:
+        """Filters a set of emoji based on another set
+        """
+        if selection_emoji_set is None:
+            # no filter on empty selection list
+            return input_emoji_set
+        filtered_emoji_set = {
+            emoji for emoji in input_emoji_set if emoji in selection_emoji_set}
+        return filtered_emoji_set
+
+    @staticmethod
+    def remove_stopwords(text_s, stopwords: List[str]) -> str:
+        """Removes a list of words from string,
+        including hyperlinks (<a></a>) and
+        integer numbers (e.g. 2012)
+        """
+        # first remove hyperlinks
+        text_s = Utils.remove_hyperlinks(text_s)
+        # split string by space character into list
+        querywords = text_s.split()
+        # clean list by matching against stopwords
+        resultwords = [word for word in querywords if word.lower()
+                       not in stopwords and not word.isdigit()]
+        s_cleaned = ' '.join(resultwords)
+        return s_cleaned
+
+    @staticmethod
+    def remove_hyperlinks(text_s):
+        """Regex to remove any hyperlinks from string
+
+        Note:
+        - anything between <a>xxx</a> will be kept
+        """
+        pattern = r'<(a|/a).*?>'
+        result = re.sub(pattern, "", text_s)
+        return result
+
+    @staticmethod
+    def _is_number(number_s):
         """Check if variable is number (float)"""
         try:
-            float(s)
+            float(number_s)
             return True
         except ValueError:
             return False
 
-    @classmethod
-    def init_main(cls):
+    @staticmethod
+    def check_intersect_polylist(
+            latlng_point, polylist, poly_exclude_list=None):
+        """Checks intersection of Point(lat, lng) against
+        list of polygons.
+
+        Arguments:
+            latlng_point {Fiona Point} -- coordinate
+            polylist {list} -- list of polys
+            poly_exclude_list {list} -- list of polys to exclude
+        """
+        include = False
+        if polylist is not None:
+            for poly in polylist:
+                if latlng_point.within(poly):
+                    include = True
+                    break
+            if include is False:
+                return False
+        else:
+            include = True
+        if poly_exclude_list is not None:
+            for poly in poly_exclude_list:
+                if latlng_point.within(poly):
+                    return False
+        if include is True:
+            return True
+        return False
+
+    @staticmethod
+    def init_main():
         """Initializing main procedure
         if package is executed directly
 
@@ -138,42 +298,65 @@ class Utils():
         """
         # set console view parameters
         # stretch console
-        os.system('mode con: cols=197 lines=40')
+        if platform.system() == 'Windows':
+            os.system('mode con: cols=197 lines=500')
         logging.getLogger("fiona.collection").disabled = True
-        cfg = BaseConfig()
-        return cfg
 
     @staticmethod
-    def _set_logger(output_folder):
+    def set_logger(output_folder: Path = None, logging_level=None):
         """ Set logging handler manually,
         so we can also print to console while logging to file
         """
+        # reset logging in case Jupyter Notebook has
+        # captured stdout
+        reload(logging)
+        # Create or get logger with specific name
+        log = logging.getLogger("tagmaps")
+        if not log:
+            # no logging handler found for tagmaps
+            # indicates package import mode
+            return
+        if log.handlers:
+            # only add log handlers once
+            return log
+        if logging_level is None:
+            logging_level = logging.INFO
         if output_folder is not None:
+            if not output_folder.exists():
+                Utils.init_dir(output_folder)
             # input(f'{type(output_folder)}')
             __log_file = output_folder / 'log.txt'
-
+        log.format = '%(message)s'  # type: ignore
+        log.datefmt = ''  # type: ignore
+        log.setLevel(logging_level)
         # Set Output to Replace in case of
         # encoding issues (console/windows)
-        sys.stdout = io.TextIOWrapper(
-            sys.stdout.detach(), sys.stdout.encoding, 'replace')
-        # Create logger with specific name
-        log = logging.getLogger("tagmaps")
-        log.format = '%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s'
-        log.datefmt = '%H:%M:%S'
-        log.setLevel(logging.DEBUG)
-        if output_folder is not None:
-            log.addHandler(logging.FileHandler(__log_file, 'w', 'utf-8'))
-        log.addHandler(logging.StreamHandler())
+        if isinstance(sys.stdout, io.TextIOWrapper):
+            # only for console output (not Juypter Notebook stream)
+            sys.stdout = io.TextIOWrapper(
+                sys.stdout.detach(), sys.stdout.encoding, 'replace')  # type: ignore
+            log.addHandler(logging.StreamHandler())
+            if output_folder is not None:
+                # only log to file in console mode
+                log.addHandler(
+                    logging.FileHandler(__log_file, 'w', 'utf-8'))  # type: ignore
+        else:
+            # log to stdout, not stderr in Jupyter Mode to prevent
+            # log.Info messages appear as red boxes
+            logging.basicConfig(
+                stream=sys.stdout, format=log.format,  # type: ignore
+                level=logging_level, datefmt=None)
+            # log.stream = sys.stdout
         # flush once to clear console
-        sys.stdout.flush()
+        # sys.stdout.flush()
         return log
 
     @staticmethod
-    def _init_output_dir(output_folder):
-        """Creates local output dir if not exists"""
-        if output_folder is not None and not output_folder.exists():
-            output_folder.mkdir()
-            print(f'Folder {output_folder.name}/ was created')
+    def init_dir(path_folder: Path):
+        """Creates local dir if not exists"""
+        if not path_folder.exists():
+            path_folder.mkdir()
+            print(f'Folder {path_folder.name}/ was created')
 
     @staticmethod
     def query_yes_no(question, default="yes"):
@@ -204,16 +387,15 @@ class Utils():
                 return valid[default]
             elif choice in valid:
                 return valid[choice]
-            else:
-                sys.stdout.write("'yes' or 'no' "
-                                 "(or 'y' or 'n').\n")
+            sys.stdout.write(
+                "'yes' or 'no' (or 'y' or 'n').\n")
 
     @staticmethod
     def daterange(start_date, end_date):
         """Return time difference between two dates"""
-        for n in range(
+        for n_val in range(
                 int((end_date - start_date).days)):
-            yield start_date + timedelta(n)
+            yield start_date + timedelta(n_val)
 
     @staticmethod
     def haversine(lon1, lat1, lon2, lat2):
@@ -227,16 +409,16 @@ class Utils():
         # haversine formula
         dlon = lon2 - lon1
         dlat = lat2 - lat1
-        a = (sin(dlat/2)**2 + cos(lat1) *
-             cos(lat2) * sin(dlon/2)**2)
-        c = 2 * asin(sqrt(a))
+        a_value = (sin(dlat/2)**2 + cos(lat1) *
+                   cos(lat2) * sin(dlon/2)**2)
+        c_value = 2 * asin(sqrt(a_value))
         # Radius of earth in kilometers is 6371
-        km = 6371 * c
-        m = km*1000
-        return m
+        km_dist = 6371 * c_value
+        m_dist = km_dist*1000
+        return m_dist
 
     @staticmethod
-    def _get_radians_from_meters(dist):
+    def get_radians_from_meters(dist):
         """Get approx. distance in radians from meters
 
         Args:
@@ -259,7 +441,7 @@ class Utils():
         return radians_dist
 
     @staticmethod
-    def _get_meters_from_radians(dist):
+    def get_meters_from_radians(dist):
         """Get approx. distance in meters from radians
         distance
 
@@ -282,7 +464,7 @@ class Utils():
         return meters_dist
 
     @staticmethod
-    def _get_emojiname(emoji_string):
+    def get_emojiname(emoji_string):
         """"Tries to get a name representation for
         emoji. Emoji can either be a single character,
         or a number of characters that construct a grapheme cluster.
@@ -295,9 +477,9 @@ class Utils():
             # if single character
             emoji_name = Utils._get_unicode_name(emoji_string)
         if not emoji_name:
-            for c in emoji_string:
+            for char_s in emoji_string:
                 emoji_name = Utils._get_unicode_name(
-                    c)
+                    char_s)
                 if emoji_name:
                     break
         if not emoji_name:
@@ -328,35 +510,20 @@ class Utils():
         # name = name(str_emoji)
         try:
             if unicodedata.name(
-                char_unicode).startswith(
-                    ("EMOJI MODIFIER",
-                     "VARIATION SELECTOR",
-                     "ZERO WIDTH")
+                    char_unicode
+            ).startswith(
+                ("EMOJI MODIFIER",
+                 "VARIATION SELECTOR",
+                 "ZERO WIDTH")
             ):
                 return False
-            else:
-                return True
+            return True
         except ValueError:
             print(char_unicode)
             return True
 
     @staticmethod
-    def split_count(text_with_emoji):
-        emoji_list = []
-        # use \X (eXtended grapheme cluster) regular expression:
-        data = regex.findall(r'\X', text_with_emoji)
-        for word in data:
-            if any(char in emoji.UNICODE_EMOJI for char in word):
-                emoji_list.append(word)
-        return emoji_list
-
-    @staticmethod
-    def extract_flags(text_with_flags):
-        flags = regex.findall(u'[\U0001F1E6-\U0001F1FF]', text_with_flags)
-        return flags
-
-    @staticmethod
-    def _extract_emoji(string_with_emoji):
+    def extract_emoji(string_with_emoji: Optional[str]) -> Set[str]:
         """Extract emoji and flags using regex package
 
         This is a new version to extract emoji (see old method:
@@ -364,7 +531,7 @@ class Utils():
         https://stackoverflow.com/a/49242754/4556479
         This method supports extracting grapheme clusters,
         emoji constructed of multiple emoji (the "perceived
-        pictograms") e.g.: 👨‍👩‍👦‍👦
+        pictograms")
 
         Compare:
         A: _extract_emoji_old:
@@ -375,12 +542,19 @@ class Utils():
         Total emoji count for the 100 most
         used emoji in selected area: 25793.
         Total distinct emoji (DEC): 1349
-        """
 
-        emoji_split = Utils.split_count(string_with_emoji)
-        emoji_list = [emoji for emoji in emoji_split]
-        flags_list = Utils.split_count(string_with_emoji)
-        return emoji_list + flags_list
+        Original Code @ sheldonzy:
+        https://stackoverflow.com/a/49242754/4556479
+        """
+        emoji_set = set()
+        if not string_with_emoji:
+            return emoji_set
+        # use \X (eXtended grapheme cluster) regular expression:
+        data = regex.findall(r'\X', string_with_emoji)
+        for grapheme in data:
+            if any(char in emoji.UNICODE_EMOJI for char in grapheme):
+                emoji_set.add(grapheme)
+        return emoji_set
 
     @staticmethod
     def _extract_emoji_old(string_with_emoji):
@@ -452,79 +626,38 @@ class Utils():
         elif str_text.lower() in (
                 'no', 'false', 'f', 'n', '0'):
             return False
-        else:
-            raise argparse.ArgumentTypeError(
-                'Boolean value expected.')
-
-    @staticmethod
-    def plot_polygon(polygon):
-        """Plot a polygon in matplotlib pyplot interface"""
-        fig = plt.figure(figsize=(10, 10))
-        ax = fig.add_subplot(111)
-        margin = .3
-        x_min, y_min, x_max, y_max = polygon.bounds
-        ax.set_xlim([x_min-margin, x_max+margin])
-        ax.set_ylim([y_min-margin, y_max+margin])
-        patch = PolygonPatch(polygon, fc='#999999',
-                             ec='#000000', fill=True,
-                             zorder=-1)
-        ax.add_patch(patch)
-        return fig
+        raise argparse.ArgumentTypeError(
+            'Boolean value expected.')
 
     @staticmethod
     def get_rectangle_bounds(points):
-        limYMin = np.min(points.T[1])
-        limYMax = np.max(points.T[1])
-        limXMin = np.min(points.T[0])
-        limXMax = np.max(points.T[0])
-        return limYMin, limYMax, limXMin, limXMax
+        """Get rectangle bounds for numpy.ndarray of point coordinates"""
+        RectangleBounds = namedtuple(
+            'RectangleBounds',
+            'lim_lat_min lim_lat_max lim_lng_min lim_lng_max')
+        lim_y_min = np.min(points.T[1])
+        lim_y_max = np.max(points.T[1])
+        lim_x_min = np.min(points.T[0])
+        lim_x_max = np.max(points.T[0])
+        return RectangleBounds(lim_y_min, lim_y_max, lim_x_min, lim_x_max)
 
     @staticmethod
-    def _get_weight(id: int,
-                    post_count: int,
-                    unique_user_count: int):
-        """Get Weight for input metrics
-        user count and post count based on
-        three types of weighting formulas (id):
-        1: weightsv1 -> Standard weighting formula
-           (x**y means x raised to the power y);
-           +1 to UserCount: prevent 1-2
-           Range from being misaligned
-        2: weightsv2 -> less importance
-           on User_Count compared
-           to photo count [Join_Count];
-           +1 to UserCount: prevent 1-2
-           Range from being misaligned
-        3: weightsv3 -> Ignores User_Count,
-           this will emphasize individual
-           and very active users
-        """
-        if id == 1:
-            weighted_value = post_count * (
-                sqrt(1/(post_count / (
-                    unique_user_count+1))**3))
-        elif id == 2:
-            weighted_value = post_count * (
-                sqrt(1/(post_count / (
-                    unique_user_count+1))**2))
-        elif id == 3:
-            weighted_value = sqrt(
-                (post_count+(2*sqrt(post_count)))*2)
-        return weighted_value
-
-    @staticmethod
-    def filter_tags(taglist: Iterable[str],
+    def filter_tags(taglist: Set[str],
                     sort_out_always_set: Set[str],
-                    sort_out_always_instr_set: Set[str]
+                    sort_out_always_instr_set: Set[str],
+                    select_tags_set: Set[str] = None
                     ) -> Tuple[Set[str], int, int]:
         """Filter list of tags based on two stoplists
 
-        Also removes numeric items and duplicates
+        - also removes numeric items and duplicates
+        - extracts only tags with len(s) > 1
+        - optionally filters list according to positive selection list
 
         Args:
             taglist (Iterable[str]): List/Set of input tags to filter
             sort_out_always_set (Set[str]): Filter complete match
             sort_out_always_instr_set (Set[str]): Filter partial match
+            select_tags_set (Set[str]): Positive filter list
 
         Returns:
             Tuple[Set[str], int, int]: Filtered list,
@@ -535,21 +668,55 @@ class Utils():
         count_tags = 0
         count_skipped = 0
 
-        photo_tags_filtered = set()
+        tags_filtered = set()
         for tag in taglist:
             count_tags += 1
-            # exclude numbers and those tags that are in sort_out_always_set
-            # or sort_out_always_instr_set
-            if (len(tag) == 1 or tag == '""'
-                or tag.isdigit()
-                    or tag in sort_out_always_set):
-                count_skipped += 1
-                continue
-            for in_str_partial in sort_out_always_instr_set:
-                if in_str_partial in tag:
+            if select_tags_set is not None:
+                # positive list available,
+                # return only selected items
+                if tag in select_tags_set:
+                    tags_filtered.add(tag)
+                else:
                     count_skipped += 1
-                    break
             else:
-                photo_tags_filtered.add(tag)
-        return (photo_tags_filtered,
+                # exclude numbers and those tags
+                # that are in sort_out_always_set
+                # or sort_out_always_instr_set
+                if (len(tag) == 1 or tag == '""'
+                        or tag.isdigit()
+                        or tag in sort_out_always_set):
+                    count_skipped += 1
+                    continue
+                for in_str_partial in sort_out_always_instr_set:
+                    if in_str_partial in tag:
+                        count_skipped += 1
+                        break
+                else:
+                    # final else Clause on loop statement
+                    tags_filtered.add(tag)
+        return (tags_filtered,
                 count_tags, count_skipped)
+
+    @staticmethod
+    def get_index_of_item(
+            l_tuple_str: List[ItemCounter], value: Optional[str]) -> int:
+        """Get index pos from list of tuples.
+
+        Stops iterating through the list as
+        soon as it finds the value
+        """
+        for pos, tuple_str in enumerate(l_tuple_str):
+            if tuple_str.name == value:
+                return pos
+        # Matches behavior of list.index
+        raise ValueError(f"{value}: x not in list")
+
+    @staticmethod
+    def get_locname(item: str, loc_name_dict: Optional[Dict[str, str]]):
+        """Gets location name from ID, if available"""
+        if loc_name_dict:
+            item_name = loc_name_dict.get(item, item)
+        else:
+            # default to item if dict key not found
+            item_name = item
+        return item_name
